(* The interactive tool to interact with the Timbuk library.
   An OCaml toplevel with tree automata functions. *)

open Tcommon

let help_string= "\nTaml functions:            

*** Creating alphabets, variable lists, terms, rewrite rules, tree automata, 
*** states, regexp (tree regular expressions), sregexp (simplified regexp).

- alphabet: (s: string) -> alphabet ---- Builds an alphabet from a string.
  For example: let alpha= alphabet \"f:2 g:1 a:0 b:0\";;

- variables: (s: string) -> variable list ---- Builds a list of variables from a
  string. For example: let vl= variables \"X Y Z\";;

- term: (a: alphabet) -> (s:string) -> term ---- Builds a (ground) term from
  an alphabet and a string.
  For example: let t= term alpha vl \"f(g(X),Y)\";;

- term_list: (a: alphabet) -> (v: variable list) -> (sl:string list) -> 
  term list ---- Builds a list of terms from an alphabet, a list of variables 
  and a list of strings. For example: 
  let tl= term_list alpha vl [\"a\";\"f(g(X),Y)\";\"f(a,a)\"];;

- rule: (a:alphabet) -> (v: variable list) -> (s: string) -> rule ---- Builds a 
  rule on alphabet a and variable list v from a string s. 
  For example: let r= rule alpha vl \"f(g(X),Y) -> g(X)\";;

- trs: (a: alphabet) -> (v: variable list) -> (sl: string list) -> rule list 
  ---- Builds a rule list on alphabet a and variable list v from a list of 
  strings. For example: 
  let r1= trs alpha vl [\"f(g(X),Y) -> g(X)\";\"a -> b\"];;

- automaton: (a: alphabet) -> (s: string) -> automaton ---- Builds an automaton
  on alphabet a from a string s.
  For example: 
  let a= automaton alpha \"States q0 q1 
                           Final States q0 
                           Transitions
                               a -> q1
                               f(q0,q1) -> q0
                               q0 -> q1\";;

- term_list_to_automaton (ts:term list) -> automaton ---- generates a tree 
  automaton whose recognized language is ts

- state: (i:int) -> state ---- Builds a state from its integer label.
  For example: let s= state 3;;

- equation: (:alphabet) -> (v: variable list) -> (s: string) -> equation ---- 
  Builds an equation on alphabet a and variable list v from a string s.
  For example: let e= equation alpha vl \"f(g(X),Y) = f(X,Z)\";;

- equation_list: (:alphabet) -> (v: variable list) -> (sl: string list) -> 
  equation list ---- Builds an equation list on alphabet a and variable list 
  v from a list of strings sl.
  For example: 
  let el= equation_list alpha vl [\"f(g(X),Y) = g(X)\";\"a = b\"];;

- regexp: (a: alphabet) (s: string) -> regexp ---- Builds a regexp 
  on alphabet a from a string s.
  For example: let r1= regexp alpha \"f(a, ((f(a, b) * b) . b (a)))\";;

- sregexp: (a: alphabet) (s: string) -> regexp ---- Builds a **simplified** 
  regexp on alphabet a from a string s.
  For example: let sr1= sregexp alp \"f(a,[f(a,*|a)])\";;


*** Tree automata functions ****************************************************

- is_recognized: (t:term) -> (a: automaton) -> (q:state) -> bool ---- Is term 
  (or automaton configuration) t recognized in state q by automaton a

- filter: (tl:term list) -> (a: automaton) -> term list ---- Filters out terms 
  of tl that do not belong to L(a), i.e. that are not recognized by a  

- gen_terms (a:automaton) -> (n:int) -> term list ---- generates the list of 
  terms, of maximal height n, recognized by a.

- determinise: automaton -> automaton ---- Determinisation of tree automata

- inter: (a1: automaton) -> (a2: automaton) -> automaton ---- Builds an
  automaton recognizing the intersection between L(a1) and L(a2).

- union: (a1: automaton) -> (a2: automaton) -> automaton ---- Builds an
  automaton recognizing L(a1) union L(a2).

- simplify: automaton -> automaton ---- Accessibility cleaning followed by 
  utility cleaning.

- unepsilon: (a: automaton) -> automaton ---- builds an equivalent tree 
  automaton without epsilon transitions, recognizing the same language. Be 
  careful, **unepsilon does not preserve R-compatibility.**

- inverse: (a:automaton) -> automaton ---- builds an automaton recognizing the 
  inverse of L(a).

- is_empty (a:automaton) -> bool ---- is the language recognized by a empty?

- is_included (a1:automaton) -> (a2:automaton) -> bool ---- is L(a1) included 
  in L(a2).

- is_equal (a1:automaton) -> (a2:automaton) -> bool ---- is L(a1) equal to 
  L(a2).
  
- term_list_to_automaton (ts:term list) -> automaton ---- generates a tree 
  automaton whose recognized language is ts

- irr: (r: rule list) -> automaton ---- Builds a tree automaton recognising the 
  set of terms irreducible by TRS r. 

- matching: (t: term) -> (a:automaton) -> (state * substitution list) list=
  matches term t on a and returns the list of couples (q,sl) where q is a state 
  and sl is a list of substitutions such that for all substitution s in sl, 
  t s -A->* q.

- add_trans (a:automaton) -> (lt: transition list) -> automaton ---- adds 
  transitions lt to tree automaton a

- merge (a:automaton) (ml: (state * state) list): automaton ---- merges all 
  pairs of states of ml in the tree automaton a

- automaton_to_string (a:automaton) -> string ---- produces the pretty printed 
  string of automaton a

- automaton_alphabet (a:automaton) -> alphabet ---- alphabet associated to a 
  given tree automaton


*** Tree Automata Completion functions *****************************************

- timbuk_completion (i:int) -> (a:automaton) -> (trs: rule list) -> 
  (eq: equation list) -> (pl: term list) -> (verbose:bool) -> 
  (automaton * bool * (term * state * substitution list) list)

  timbuk's completion: performs a maximum of i completion steps with trs and 
  simplification with eq and checks if a term of pl list is reachable. In the 
  result, the 'automaton' is the completed automaton, the 'bool' is true if the 
  automaton is complete (R-compatible), the '(term * state * substitution list) 
  list' is the list of all substitutions found when matching terms of pl on the 
  completed tree automaton. If this list is empty, no pattern of pl is reachable.

- timbuk_fun_completion (i:int) -> (a:automaton) -> (atc: automaton) -> 
  (trs: rule list) -> (pl: term list) -> (equation list * automaton * bool)

  timbuk's completion for functional TRS: performs a maximum of i completion 
  steps with trs. trs is supposed to be terminating and sufficiently complete 
  (i.e. all rewriting derivation eventually leads to a well typed term of T(C).
  Simplication equations are automatically generated from the automaton atc 
  which recognizes well typed constructor terms of T(C). The set of equation is 
  generated so that terms of pl should not be reachable (equations are 
  automatically built and refined). In the result, the 'equation list' is the 
  list of automatically generated equations to achieve a terminating completion,
  the 'automaton' is the completed automaton, the 'bool' is true if the 
  automaton is complete (R-compatible). 


*** Term functions *************************************************************

- apply: (s:substitution) -> (t:term) -> term ---- applies substitution s on 
  term t

- normalize_term (t: term) -> (a:automaton) -> (q:state): transition list ---- 
  gives the necessary transitions to add to a for term t to be recognized in 
  state q


*** Substitution functions *****************************************************

- apply: (s:substitution) -> (t:term) -> term ---- applies substitution s on 
  term t

- value: (s:substitution) -> (v:variable) -> state option ---- If s is a state 
  substitution, returns the value Some(q) if the state associated to v in s is 
  q, and None otherwise

- domain: (s:substitution) -> variable list ---- the domain of a substitution


*** Rewrite Rules functions ****************************************************

- lhs: (r: rule) -> term ---- Gets the left-hand side of a rewrite rule

- rhs: (r: rule) -> term ---- Gets the right-hand side of a rewrite rule

- conditions: (r: rule) -> conditions list ---- Gets the condition list  of a 
  rewrite rule (if any)

- rule_alphabet (r:rule) -> alphabet ---- alphabet associated to a given 
  rewrite rule


*** Equation accessors *********************************************************

- elhs: (r: equation) -> term ---- Gets the left-hand side of an equation

- erhs: (r: equation) -> term ---- Gets the right-hand side of an equation


*** Regular expressions functions **********************************************

- regexp_to_automaton (r:regexp) -> automaton ---- builds a tree automaton from 
  the regular expression r

- sregexp_to_automaton (sr:sregexp) -> automaton ---- builds a tree automaton 
  from the simplified regular expression sr

- sregexp_to_regexp (sr:sregexp) -> regexp ---- converts a simplified regular 
  sr expression to a regular expression

- automaton_to_regexp (aut:automaton) -> regexp ---- converts an automaton aut  
  to a regular expression

- automaton_to_sregexp (aut:automaton) -> sregexp ---- converts an automaton to 
  a **simplified** regular expression if this is possible. Otherwise a 
  ConversionFailure exception is raised

- regexp_alphabet (r:regexp) -> alphabet ---- alphabet associated to a given 
  regexp

- sregexp_alphabet (s:sregexp) -> alphabet ---- alphabet associated to a given 
  sregexp


*** The help function **********************************************************

- help: unit -> unit ---- Prints this help message

";;


let help()= print_string help_string

(* borrowed from new ocaml string lib *)
let split_on_char (sep:char) (s:string) =
  let r = ref [] in
  let j = ref (String.length s) in
  for i = String.length s - 1 downto 0 do
    if String.unsafe_get s i = sep then begin
      r := String.sub s (i + 1) (!j - i - 1) :: !r;
      j := i
    end
  done;
  String.sub s 0 !j :: !r

let help_html()= print_string (String.concat "<br>\n" (split_on_char '\n' help_string))

(* ----- TYPE ALIASING ------ for unified displaying in the Taml loop *) 
type variable=string 
type alphabet=Alphabet.t

(* In the taml loop we associate alphabet and variable list to TRSs, and alphabet to TA, regexp etc. *)
(* type trs= (alphabet * Rewriting.trs) *)
type rule= (alphabet * Rewriting.rule)
type equation= (alphabet * Equation.equation)

let tamltrs_to_trs= (List.map snd)
type automaton= (alphabet * Automaton.automaton)
type regexp= (alphabet * Tregexp.t)
type sregexp= (alphabet * Tregexp.simple)
type substitution= Substitution.Substitution.t
type term= (alphabet * Term.t)
type state= Tcommon.state
type transition= Transitions.external_trans


(* ----- PRETTY PRINTERS ---------- of the taml loop *)
(* Automatic pretty printers *)
let variables_printer ff s= Format.fprintf ff "%s@." (String.concat " " s)
let alphabet_printer ff a= Format.fprintf ff "%s@." (Alphabet.to_string a)
(* let state_set_printer ff s= Format.fprintf ff "%s@." (State_set.to_string s) 
let term_printer ff t= Format.fprintf ff "%s@." (Term.to_string t) *)
(* let trs_printer ff (_,_,trs)= Format.fprintf ff "%s@." (Rewriting.to_string trs) *)
let automaton_printer ff (_,a) = Format.fprintf ff "%s@." (Automaton.to_string a)
let regexp_printer ff (_,a)= Format.fprintf ff "%s@." (Tregexp.to_string a)
let sregexp_printer ff (_,a)= Format.fprintf ff "%s@." (Tregexp.simple_to_string a)
let term_printer ff (_,t)= Format.fprintf ff "%s@." (Term.to_string t)
let substitution_printer ff s= Format.fprintf ff "%s@." (Substitution.Substitution.to_string s)
let state_printer ff s= Format.fprintf ff "%s@." (State.to_string s)
let rule_printer ff (_,rl)= Format.fprintf ff "%s@." (Rewriting.rule_to_string rl)
let equation_printer ff ((_,eq):equation)= Format.fprintf ff "%s@." (Equation.equation_to_string eq)
let transition_printer ff trans= Format.fprintf ff "%s@." (Transitions.external_to_string trans) 

(* We need those printers because default list printer adds \n between all elements of a list *)
let trs_printer ff rl= Format.fprintf ff "%s@." ("["^(String.concat "; " (List.map (function (_,r) -> (Rewriting.rule_to_string r)) rl))^"]\n")

(* We need those printers because default list printer adds \n between all elements of a list *)
let equation_list_printer ff (rl:equation list)= Format.fprintf ff "%s@." ("["^(String.concat "; " (List.map (function (_,r) -> (Equation.equation_to_string r)) rl))^"]\n")

(* We need those printers because default pair printer adds \n between all elements of a pair *)
let substitutionlist_to_string sl=
  ("["^(String.concat "; " (List.map (function sub -> (Substitution.Substitution.to_string sub)) sl))^"]\n")

let substitutionlist_printer ff sl= 
  Format.fprintf ff "%s@." (substitutionlist_to_string sl)

(* We need those printers because default pair printer adds \n between all elements of a pair *)
let state_substitutionlist_to_string (state,sl)= 
  ("("^(State.to_string state)^","^(substitutionlist_to_string sl)^")\n")

let state_substitutionlist_printer ff (state,sl)= Format.fprintf ff "%s@."(state_substitutionlist_to_string (state,sl))

let automaton_to_string (_,a)= Automaton.to_string a

(* We need those printers because default list printer adds \n between all elements of a list *)
let transition_list_to_string tl=
  ("["^(String.concat "; " (List.map (function trans -> (Transitions.external_to_string trans)) tl))^"]\n")

let transition_list_printer ff ltrans= Format.fprintf ff "%s@." (transition_list_to_string ltrans) 

(* specific term list printer for Taml... so that parsing and pretty printing are coherent *)

let termlist_to_string (tl: term list):string=
  ("["^(String.concat "; " (List.map (function (_,t) -> (Term.to_string t)) tl))^"]\n")

let termlist_printer ff (_,tl)= Format.fprintf ff "%s@." (termlist_to_string tl)

(* We need those printers because default pair printer adds \n between all elements of a pair *)
let state_pair_printer ff ((q1,q2):(state*state))= Format.fprintf ff "%s@." ("("^(State.to_string q1)^","^(State.to_string q2)^")")

(* this one is not coherent with the parsing format:
   let termset_printer ff (_,ts)= Format.fprintf ff "%s@." (Automaton.Termset.to_string ts) *)


(* let specification_printer ff s= Format.fprintf ff "%s@." (Specification.to_string s) *)

(* -----MAIN TAML FUNCTIONS ------------- *) 

(* Main Taml Functions. See comments in the help() function above *)

(* reinitialize the parser error counters *)
let parser_reinit()= ErrorHandler.reinit ()

let alphabet (s:string):alphabet= 
  (* parseErrorHandling admits only function and one parameter. We have to wrap the parsing code
     so as to satisfy this pattern *)
  let f (lexer,stream)=
    parser_reinit();
    Alphabet.from_list (Tparser.opslist lexer stream)
  in
  Dataextracting.parseErrorHandling f (Tlexer.token, (Lexing.from_string s))

let variables (s:string):variable list= 
  (* parseErrorHandling admits only function and one parameter. We have to wrap the parsing code
     so as to satisfy this pattern *)
  let f (lexer,stream)=
    parser_reinit();
    Tparser.variablesList lexer stream
  in Dataextracting.parseErrorHandling f (Tlexer.token, (Lexing.from_string s))
      
let term (a:alphabet) (v:variable list) (s:string):term= 
  (* parseErrorHandling admits only function and one parameter. We have to wrap the parsing code
     so as to satisfy this pattern *)
  let f (lexer,stream)=
    parser_reinit();
    Tglobal.alpha:= Alphabet.to_list a;  (* We initialize the alphabet used for parsing *)
    Tglobal.varset:= v;  
    let t= (Tparser.term lexer stream) in
    (a, (Term.pTerm_to_term t))
  in Dataextracting.parseErrorHandling f (Tlexer.token, (Lexing.from_string s))

let rec term_list (a:alphabet) (v:variable list) (sl:string list):term list= 
  match sl with
      [] -> []
    | s::rem -> (term a v s)::(term_list a v rem)

(* let trs (a:alphabet) (v: variable list) (s:string):trs= 
  (* parseErrorHandling admits only function and one parameter. We have to wrap the parsing code
     so as to satisfy this pattern *)
  let f (lexer,stream)=
    parser_reinit();
    Tglobal.alpha:= Alphabet.to_list a;  (* We initialize the alphabet used for parsing *)
    Tglobal.varset:= v;                  (* We initialize the list of variables used for parsing *)
    (a,Rewriting.trs_from_parser_format (Tparser.trsRulesList lexer stream))
  in Dataextracting.parseErrorHandling f (Tlexer.token, (Lexing.from_string s))
*)

let rule (a:alphabet) (v: variable list) (s:string):rule= 
  (* parseErrorHandling admits only function and one parameter. We have to wrap the parsing code
     so as to satisfy this pattern *)
  let f (lexer,stream)=
    parser_reinit();
    Tglobal.alpha:= Alphabet.to_list a;  (* We initialize the alphabet used for parsing *)
    Tglobal.varset:= v;                  (* We initialize the list of variables used for parsing *)
    (a,Rewriting.rule_from_parser_format (Tparser.trsrule lexer stream))
  in Dataextracting.parseErrorHandling f (Tlexer.token, (Lexing.from_string s))

let lhs ((alpha,(l,r,cl):rule)): term= (alpha,l)
let rhs ((alpha,(l,r,cl):rule)): term= (alpha,r)
let conditions ((alpha,(l,r,cl):rule))= cl

let rec trs (a:alphabet) (v: variable list) (sl: string list):rule list= 
  match sl with
      [] -> []
    | s::rem -> (rule a v s)::(trs a v rem)

let equation (a:alphabet) (v:variable list) (s:string):equation=
   (* parseErrorHandling admits only function and one parameter. We have to wrap the parsing code
     so as to satisfy this pattern *)
  let f (lexer,stream)=
    parser_reinit();
    Tglobal.alpha:= Alphabet.to_list a;  (* We initialize the alphabet used for parsing *)
    Tglobal.varset:= v;                  (* We initialize the list of variables used for parsing *)
    let (t1,t2)= (Tparser.simpleEq lexer stream) in
    (a,Equation.Regular(Term.pTerm_to_term t1,Term.pTerm_to_term t2,[]))
  in Dataextracting.parseErrorHandling f (Tlexer.token, (Lexing.from_string s))

let rec equation_list (a:alphabet) (v: variable list) (sl: string list):equation list= 
  match sl with
      [] -> []
    | s::rem -> (equation a v s)::(equation_list a v rem)

let elhs((alpha,eq):equation): term=
  match eq with
      Equation.Regular(l,_,_) -> (alpha,l) 
    | Equation.Exteq(_,_,_) | Equation.Extany(_,_,_) -> failwith ("Taml: elhs not supported for those equations")

let erhs((alpha,eq):equation): term=
  match eq with
      Equation.Regular(_,r,_) -> (alpha,r) 
    | Equation.Exteq(_,_,_) | Equation.Extany(_,_,_) -> failwith ("Taml: erhs not supported for those equations")


let automaton (a:alphabet) (s:string):automaton= 
(* parseErrorHandling admits only function and one parameter. We have to wrap the parsing code
     so as to satisfy this pattern *)
  let f (lexer,stream)=
    parser_reinit();
    Tglobal.alpha:= Alphabet.to_list a;  (* We initialize the alphabet used for parsing *)
    (a,snd (Dataextracting.automaton_construction (Tparser.automatonTopLevel lexer stream)))
  in Dataextracting.parseErrorHandling f (Tlexer.token, (Lexing.from_string s))

let state (i:int): state= i

let check_alpha_term_list (alpha1:alphabet) (tl:term list):bool=
  List.for_all (function (alpha2,term) -> alpha1=alpha2) tl

let coherent_alpha (l: (alphabet*'a) list):bool=
  if l=[] then true
  else let alpha1=fst (List.hd l) in
       List.for_all (function (alpha2,_) -> alpha1=alpha2) (List.tl l)

let filter (tl: term list) ((alpha1,aut):automaton): term list=
  if not(check_alpha_term_list alpha1 tl) then failwith "Incoherence between the alphabets of the automaton and of the term list"
  else
    List.filter (function (_,term) -> (Automaton.is_recognized aut term)) tl

let regexp (a:alphabet) (s:string):regexp= 
  (* parseErrorHandling admits only function and one parameter. We have to wrap the parsing code
     so as to satisfy this pattern *)
  let f (lexer,stream)=
    parser_reinit();
    Tglobal.alpha:= Alphabet.to_list a;  (* We initialize the alphabet used for parsing *)
    (a,(Tparser.regexp lexer stream))
  in Dataextracting.parseErrorHandling f (Tlexer.token, (Lexing.from_string s))

let sregexp (a:alphabet) (s:string):sregexp= 
  (* parseErrorHandling admits only function and one parameter. We have to wrap the parsing code
     so as to satisfy this pattern *)
  let f (lexer,stream)=
    parser_reinit();
    Tglobal.alpha:= Alphabet.to_list a;  (* We initialize the alphabet used for parsing *)
    (a,(Tparser.sregexp lexer stream))
  in Dataextracting.parseErrorHandling f (Tlexer.token, (Lexing.from_string s))


(* With regexp the alphabet is completed with additionnal box constants *)
let sregexp_to_regexp ((a,sr):sregexp):regexp= 
  let (r,additionnal_constants)= (Tregexp.of_simple sr) in
  (* We add all the new constants introduced by conversion to the alphabet *)
  let new_alpha= Alphabet.copy a in
  let ()= List.iter (function sym -> try Alphabet.add sym 0 new_alpha with Alphabet.Multiply_defined_symbol(_) -> ()) additionnal_constants in
  if (additionnal_constants<>[]) then print_string ("\nWarning! the alphabet is now: "^(Alphabet.to_string new_alpha)^"\n\n");
  (new_alpha,r)


(* When going back from regexp to automata, we remove the box constants in alphabet *)
let remove_box_constants (a:Alphabet.t):alphabet=
  Alphabet.from_list (List.filter (function (s,_) -> not(Tregexp.isBox s)) (Alphabet.to_list a))

let regexp_to_automaton ((a,r):regexp):automaton= 
  (remove_box_constants a,(RegexpToAutomaton.build_automaton (Alphabet.to_list a) r))

let sregexp_to_automaton ((a,sr):sregexp):automaton= 
  let (r,additionnal_constants)= (Tregexp.of_simple sr) in
  let alpha= (Alphabet.to_list a)@(List.map (function x -> (x,0)) additionnal_constants) in
  (remove_box_constants a,(RegexpToAutomaton.build_automaton alpha r))


let automaton_to_regexp ((alpha,aut):automaton):regexp=
  (* To build the regexp, we need to determinise and unepsilon the automaton. The determinisation also do the unepsi *)
  let regexp= AutomatonToRegexp.automaton_to_regexp (Alphabet.to_list alpha) (Automaton.simplify (Automaton.determinise aut)) in
  (* we add all the new constants to the alphabet *)
  let new_alpha= Alphabet.copy alpha in
  let additionnal_constants= Tregexp.allBoxes regexp in
  let ()= List.iter (function sym -> try Alphabet.add sym 0 new_alpha with Alphabet.Multiply_defined_symbol(_) -> ()) additionnal_constants in
  (new_alpha,regexp)

let automaton_to_sregexp ((alpha,aut):automaton):sregexp=
  (* To build the regexp, we need to determinise and unepsilon the automaton. The determinisation also do the unepsi *)
  (alpha,AutomatonToRegexp.automaton_to_sregexp (Alphabet.to_list alpha) (Automaton.simplify (Automaton.determinise aut))) 

let determinise (alpha,a)= (alpha,Automaton.determinise a)

let inter (alpha1,a1) (alpha2,a2): automaton= 
  if alpha1=alpha2 then (alpha1,Automaton.intersection a1 a2)
  else failwith "Incoherence between the alphabets of the two tree automata"

let union (alpha1,a1) (alpha2,a2): automaton= 
  if alpha1=alpha2 then (alpha1,Automaton.union a1 a2)
  else failwith "Incoherence between the alphabets of the two tree automata"

let unepsilon (alpha,a): automaton= (alpha,Automaton.unepsilon a)

let inverse (alpha,a):automaton= (alpha, Automaton.inverse alpha a)

let rule_alphabet ((alpha,_):rule):alphabet =alpha
let automaton_alphabet ((alpha,_):automaton):alphabet =alpha
let regexp_alphabet ((alpha,_):regexp): alphabet =alpha
let sregexp_alphabet ((alpha,_):sregexp): alphabet =alpha

let is_included ((alpha1,aut1):automaton) ((alpha2,aut2):automaton):bool=
  if alpha1=alpha2 then Automaton.is_included alpha1 aut1 aut2
  else failwith "Incoherence between the alphabets of the two tree automata"

let is_equal ((alpha1,aut1):automaton) ((alpha2,aut2):automaton):bool=
  if alpha1=alpha2 then Automaton.is_equal alpha1 aut1 aut2
  else failwith "Incoherence between the alphabets of the two tree automata"

let is_empty ((_,aut):automaton):bool= Automaton.is_empty aut

let simplify (alpha,a):automaton= (alpha, Automaton.simplify a)

let gen_terms ((alpha,aut):automaton) (n:int): term list= 
  (List.map (function term -> (alpha,term)) (Automaton.Termset.elements (Automaton.terms_of_height (Automaton.unepsilon aut) n)))

let termlist_to_automaton (tl:term list): automaton=
  if not(coherent_alpha tl) then failwith "Incoherence between the alphabets of the terms in the list"
  else
    if tl=[] then failwith "Cannot generate an automaton for an empty list of terms: no alphabet available"
    else 
      let alpha= fst (List.hd tl) in
      (* remove the alphabet info *)
      let tl2= List.map snd tl in
      let (aut,_,_)= Automaton.term_to_automaton tl2 [] (Automaton.new_s2t ()) in
      (alpha,aut)

let irr (alpha:alphabet) (trs: rule list):automaton= 
  let trs2= (tamltrs_to_trs trs) in
  let variables= Rewriting.vars_trs trs2 in
  (alpha,Automaton.build_irr_opt alpha variables trs2)

let matching ((alpha1,t):term) ((alpha2,aut):automaton): (state * substitution list) list=
  if alpha1<>alpha2 then failwith ("Matching impossible alphabet of term and automaton do not coincide!")
  else
    match t with 
        (* Matching of variable terms is not necessary in timbuk completion algo. Thus we implement it here 
           solution is the following: a variable X matches all states q and corresponding subst is {X -> q} *)
        Term.Var(s) -> 
          let all_states= Automaton.get_states aut in
          List.map (function state -> (state, [(Substitution.Substitution.add s (Term.State(state)) Substitution.Substitution.empty)])) all_states
      | Term.Fsym(_,_) | Term.State(_) -> 
          let variables= Term.allVars t in
          List.map (function (_,s,substs) -> (s,Substitution.SubstitutionSet.elements substs)) (Completion.check_intersection aut [t] variables)

let apply (s: substitution) ((alpha,t):term): term=
  (alpha,Rewriting.subst_apply s t)

type 'a option= None | Some of 'a

let value (s: substitution) (v:variable):  state option=
  try (
    match (Substitution.Substitution.find v s) with
        Term.State(q) -> Some(q)
      | Term.Var(_) | Term.Fsym(_) -> failwith ("This function does not cover substitution mapping terms to variables"))
  with Not_found -> None
  
let domain (s:substitution): variable list= Substitution.Substitution.domain s

let is_recognized ((alpha1,t):term) ((alpha2,a): automaton) (q:state):bool=
  if alpha1<>alpha2 then failwith ("Matching impossible alphabet of term and automaton do not coincide!")
  else
    match t with
        Term.State(q1) -> (* We extend recognition to state values of t.
                        if t=State(q1) and q1 -epsilon->* q then it is recognized *)
                     Transitions.is_pred q1 q a.Automaton.trans
      | Term.Var(_) | Term.Fsym(_,_) -> Automaton.is_recognized_by a t q

(* WARNING: This implementation makes a non-efficient usage of highly optimized Automata Transitions functions.
   This function is here only for pedagogical reasons *)

let normalize_term ((alpha1,t):term) ((alpha2,a):automaton) (q:state): transition list=
  if alpha1<>alpha2 then failwith ("Matching impossible: alphabet of term and automaton do not coincide!")
  else
    let delta= Transitions.copy a.Automaton.trans in
    (* normalization directly adds transitions to the tree automaton, this is avoided here
       (for pedagogical reasons) using an intermediate copy of transition set *)
    let (q2,ltrans)=(Automaton.normalize_term t delta false) in
    if q=q2 then ltrans
    else (Transitions.Epsilon(q2,q))::ltrans


(* WARNING: This implementation makes a non-efficient usage of highly optimized Automata Transitions functions.
   This function is here only for pedagogical reasons *)

let add_trans ((alpha,a):automaton) (lt: transition list): automaton=
  let delta= Transitions.copy a.Automaton.trans in 
  let fs= a.Automaton.fstates in 
  let all_new_states= Transitions.states_of_trans lt in
  (* We add new states to delta *)
  Transitions.Stateset.iter (function elt -> Transitions.add_state delta elt) all_new_states;
  List.iter (function trans -> (ignore (Transitions.add trans delta))) lt;
  (alpha,{Automaton.trans=delta;Automaton.fstates=fs})

let merge ((alpha,a):automaton) (merge_list: (state * state) list): automaton=
  let new_aut= Automaton.copy a in
  let (merged_aut,_)= Automaton.merge new_aut (Transitions.new_merge_table ()) merge_list in
  (alpha,merged_aut)


let timbuk_completion (n:int) ((alpha1,a):automaton) (trs: rule list) (eq: equation list) (pl: term list) (verbose:bool):
    (automaton * bool * (term * state * substitution list) list)=
  (* (option:completion_option list)= *)
  if not (List.for_all (function (alpha2,_) -> (alpha1=alpha2)) trs) then failwith ("Completion is impossible: alphabet of automaton and TRS do not coincide!");
  if not (List.for_all (function (alpha2,_) -> (alpha1=alpha2)) eq) then failwith ("Completion is impossible: alphabet of automaton and list of equations do not coincide!");
  Tcommon.completion_reinit_options ();
  Tcommon.exact_completion:=true;
  if verbose then Tcommon.verbose:=2;
  let messageOn (s:string)= print_string s in
  (* we remove the alphabets from trs, pl and eq. They are unecessary for external_call_to_completion *)
  let trs2= List.map snd trs in
  let pl2= List.map snd pl in
  let eq2= List.map snd eq in
  let (b,comp,_,lsub)= Completion.external_call_to_completion alpha1 trs2 a eq2 pl2 n messageOn in
  ((alpha1,comp),b,List.map (function (t,q,subSet) -> ((alpha1,t),q,Substitution.SubstitutionSet.elements subSet)) lsub)


let timbuk_fun_completion (n:int) ((alpha1,a):automaton) ((alphatc,atc):automaton) (trs: rule list) (pl: term list):
    (equation list * automaton * bool)=
  if not (Alphabet.is_included alphatc alpha1) then failwith ("Completion is impossible: alphabet of automaton and TRS do not coincide!");
  if not (List.for_all (function (alpha2,_) -> (alpha1=alpha2)) trs) then failwith ("Completion is impossible: alphabet of automaton and TRS do not coincide!");
  Tcommon.completion_reinit_options ();
  Tcommon.fun_verif:=true;
  (* we remove the alphabets from trs, pl and eq. They are unecessary for extVerifFun *)
  let trs2= List.map snd trs in
  let pl2= List.map snd pl in
  let trs_pl_vars= List.fold_right 
    (function t -> function l ->
      Tcommon.union (Term.vars t) l)
    pl2 (Rewriting.vars_trs trs2) in 
  let alpha1copy= Alphabet.copy alpha1 in
  try(
    let (b,comp,el,i,_)= Verification.extVerifFun alpha1copy trs_pl_vars trs2 a atc pl2 n in
    (* we need to project comp automaton on symbols of the alphabet, because there 
       are some additionnal symbols used for the encoding of TRS into grammars (#q symbols e.g.) *)
    let alpha_symb= List.map fst (Alphabet.to_list alphatc) in
    let alpha_comp= (Automaton.project comp alpha_symb) in
(*    (* we need to project alpha_comp automaton on constructor symbols (could have also used intersection instead of projection) *)
    let consts_symb= List.map fst (Alphabet.to_list (Automaton.transitions_alphabet atc)) in
    let tc_comp= (Automaton.project (Automaton.unepsilon alpha_comp) consts_symb) in *)
    ((List.map (function e -> (alpha1,e)) el),(alpha1,alpha_comp),b))
  with Equation_generation.ExhaustCSN -> print_string ("The maximal number of completion steps you gave ("^(string_of_int n)^") has been exhausted by equation generation. Increase it to succeed.\n"); failwith("Completion stopped")
    | Verification.Counterex(s) -> print_string s; failwith("Completion stopped")


let _= print_string "
 Taml:  Tree Automata for OCaml             Type 'help();;' 
                                            for help on taml functions\n"
